/*
 * Copyright (C) 2002 Brent N. Chun <bnc@caltech.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>
#include <e/bitmask.h>
#include <e/bytes.h>
#include <e/e_error.h>
#include <e/net.h>
#include <e/token_array.h>
#include <e/xmalloc.h>
#include "gexec.h"
#include "gexec_lib.h"
#include "gexec_options.h"

extern char **environ;

static void print_bad_nodes(int *bad_nodes, char **ips)
{
    char    host[1024];
    char    *ip; 
    int     bad_node;
    int     bad_nodes_copy[GEXEC_MAX_NPROCS];

    bitmask_copy(bad_nodes_copy, bad_nodes, GEXEC_MAX_NPROCS);
    while ((bad_node = bitmask_ffs(bad_nodes_copy, GEXEC_MAX_NPROCS)) != -1) {
        ip = ips[bad_node];
        net_iptohost(ip, host, 1024);
        printf("Could not connect to %s (%s)\n", host, ip);
        bitmask_clr(bad_nodes_copy, GEXEC_MAX_NPROCS, bad_node);
    }
}

static void ips_create(char ***ips, char **hosts, int nhosts)
{
    int i;

    *ips = (char **)xmalloc(nhosts * sizeof(char *));
    for (i = 0; i < nhosts; i++) {
        (*ips)[i] = (char *)xmalloc(IP_STRLEN);
        if (net_hosttoip(hosts[i], (*ips)[i]) != E_OK) {
            fprintf(stderr, "Bad host %s\n", hosts[i]);
            exit(2);
        }
    }
}

static void ips_destroy(char **ips, int nhosts)
{
    int i;
    
    for (i = 0; i < nhosts; i++)
        xfree(ips[i]);
    xfree(ips);
}

static char *malloc_fullpath(char *path)
{
    char        *fullpath = NULL;
    char        *path_env;
    char        **dirs;
    int         ndirs, i, maxdir, found_path = 0;
    struct stat stat_buf;
 
    if (path[0] == '/') {
        if (stat(path, &stat_buf) < 0) {
            fprintf(stderr, "Bad filename %s\n", path);
            exit(3);
        }
        fullpath = (char *)xmalloc(strlen(path) + NULLBYTE_SIZE);
        strcpy(fullpath, path);
        return fullpath;
    }
    if ((path_env = getenv("PATH")) == NULL) {
        fprintf(stderr, "PATH not set\n");
        exit(3);
    }
    token_array_create(&dirs, &ndirs, path_env, ':');
    maxdir   = strlen(token_array_maxtoken(dirs, ndirs));
    fullpath = xmalloc(maxdir + SLASHBYTE_SIZE + strlen(path) + NULLBYTE_SIZE);
    for (i = 0; i < ndirs; i++) {
        sprintf(fullpath, "%s/%s", dirs[i], path);
        if (stat(fullpath, &stat_buf) == 0 &&
            (stat_buf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))) {
            found_path = 1;
            break;
        }
    }
    token_array_destroy(dirs, ndirs);
    if (!found_path) {
        fprintf(stderr, "Bad filename %s\n", path);
        exit(3);
    }
    return fullpath;
}

static void gargs_create(int argc, char **argv, int *gargc, char ***gargv, int args_index)
{
    int i, len;

    *gargc = argc - args_index;
    *gargv = (char **)xmalloc((*gargc + 1) * sizeof(char *));
    (*gargv)[0] = malloc_fullpath(argv[args_index]);
    for (i = 1; i < *gargc; i++) {
        len = strlen(argv[args_index + i]);
        (*gargv)[i] = xmalloc(len + NULLBYTE_SIZE);
        strcpy((*gargv)[i], argv[args_index + i]);
    }
    (*gargv)[i] = NULL;
}

static void gargs_destroy(int gargc, char **gargv)
{
    int i;

    for (i = 0; i < gargc; i++)
        xfree(gargv[i]);
    xfree(gargv);
}

int main(int argc, char **argv)
{
    gexec_options opts;
    int           bad_nodes[GEXEC_BITMASK_SIZE];
    int           gargc, args_index, nhosts;
    char          *gexec_svrs;
    char          **hosts, **ips, **gargv;
    
    gexec_init();

    gexec_options_init(&opts);
    gexec_options_parse(&opts, argc, argv, &args_index);

    gexec_svrs = getenv("GEXEC_SVRS");
    if (gexec_svrs == NULL) {
        fprintf(stderr, "GEXEC_SVRS not set\n");
        exit(3);
    }
    token_array_create(&hosts, &nhosts, gexec_svrs, ' ');
    ips_create(&ips, hosts, nhosts);
    gargs_create(argc, argv, &gargc, &gargv, args_index);
    if (opts.nprocs == 0)
        opts.nprocs = nhosts;
    if (opts.nprocs > nhosts) {
        fprintf(stderr, "Not enough hosts in GEXEC_SVRS\n");
        exit(3);
    }
    gexec_spawn(opts.nprocs, ips, gargc, gargv, environ, 
                &opts.spawn_opts, bad_nodes);
    print_bad_nodes(bad_nodes, ips);
    gargs_destroy(gargc, gargv);
    ips_destroy(ips, nhosts);
    token_array_destroy(hosts, nhosts);

    gexec_shutdown();

    return 0;
}
